Inductive Even : nat -> Prop :=
  | Even_O : (Even O)
  | Even_S : (n:nat)(Odd n) -> (Even (S n))
with Odd : nat -> Prop :=
  | Odd_S : (n:nat)(Even n) -> (Odd (S n)).

Hints Resolve Even_O Even_S Odd_S.

Lemma even_odd : 
  (n:nat) { b:bool | (b=true /\ (Even n)) \/ (b=false /\ (Odd n)) }.
Induction n.
Exists true. Auto. 
Intro n0. Induction 1. Intro b. Case b.
  Intro Hb. Exists false. Decompose [and or] Hb. 
    Right. Auto. Discriminate H0.
  Intro Hb. Exists true. Decompose [and or] Hb.
    Discriminate H0. Left. Auto. 
Save.

Require Extraction.

Write Caml File "even_odd" [even_odd].

Write Haskell File "even_odd" [even_odd].

Lemma even_odd_prog : 
  (n:nat) { b:bool | (b=true /\ (Even n)) \/ (b=false /\ (Odd n)) }.
Realizer Fix prog { prog [n:nat] : bool :=
           Cases n of
	     O => true
           | (S p) => if (prog p) then false else true
	   end }.
Program_all.
Decompose [and or] o. Right. Auto. Discriminate H.
Decompose [and or] o. Discriminate H. Left. Auto.
Save.

Write Caml File "even_odd_prog" [even_odd_prog].

Require Programs.
Require Omega.

Correctness even_odd_rec
  let rec even (n:nat) : bool { variant n for lt } =
    (if (zerop_bool n) then
       true
     else
       let b = (even (pred n)) in 
       if b then false else true)
    { if result then (Even n) else (Odd n) }.
Proof.
Rewrite Test2. Auto.
Omega.
Rewrite Test1 in Post6.
Replace n0 with (S (pred n0)). Auto. Omega.
Rewrite Test1 in Post6.
Replace n0 with (S (pred n0)). Auto. Omega.
Save.

Require ProgramsExtraction.

Write Caml File "even_odd_rec" [even_odd_rec].

Global Variable n : nat ref.

Correctness even_odd_imp
    (let b = ref true in
     begin
       while (notzerop_bool !n) do
         { invariant 
	     (le n n@0) /\
             ((b=true/\(Even (minus n@0 n))) \/ (b=false/\(Odd (minus n@0 n))))
           variant n for lt }
         n := (pred !n);
         if !b then b := false else b := true
       done;
       !b
     end)
    { (result=true/\(Even n@)) \/ (result=false/\(Odd n@)) }.
Proof.
Split. Omega. Split. Omega.
Right. Split. Auto.
Replace (minus n (pred n0)) with (S (minus n n0)).
Decompose [and or] Pre2. Auto. Rewrite Test1 in H1. Discriminate H1.
Omega.
Split. Omega. Split. Omega.
Left. Split. Auto.
Replace (minus n (pred n0)) with (S (minus n n0)).
Decompose [and or] Pre2. Rewrite Test1 in H1. Discriminate H1. Auto.
Omega.
Split. Auto. Left. Split. Auto. Rewrite <- minus_n_n. Auto.
Decompose [and or] Post5. Left. Split. Assumption.  
Replace n with (minus n n0). Assumption. Omega.
Right. Split. Assumption. Replace n with (minus n n0). Assumption. Omega.
Save.

Initialize n with (17).
Write Caml File "even_odd_imp" [even_odd_imp].
